@@ -6,8 +6,9 @@
 //! `TorClient::connect()`.
 use crate::address::IntoTorAddr;
 
-use crate::config::{ClientAddrConfig, TorClientConfig};
-use tor_circmgr::{IsolationToken, StreamIsolationBuilder, TargetPort};
+use crate::config::{ClientAddrConfig, StreamTimeoutConfig, TorClientConfig};
+use tor_circmgr::{DirInfo, IsolationToken, StreamIsolationBuilder, TargetPort};
+use tor_config::MutCfg;
 use tor_dirmgr::DirEvent;
 use tor_persist::{FsStateMgr, StateMgr};
 use tor_proto::circuit::ClientCirc;
@@ -18,7 +19,7 @@ use futures::stream::StreamExt;
 use futures::task::SpawnExt;
 use std::convert::TryInto;
 use std::net::IpAddr;
-use std::sync::{Arc, Weak};
+use std::sync::{Arc, Mutex, Weak};
 use std::time::Duration;
 
 use crate::{Error, Result};
@@ -32,6 +33,9 @@ use tracing::{debug, error, info, warn};
 /// Cloning this object makes a new reference to the same underlying
 /// handles: it's usually better to clone the `TorClient` than it is to
 /// create a new one.
+// TODO(nickm): This type now has 5 Arcs inside it, and 2 types that have
+// implicit Arcs inside them! maybe it's time to replace much of the insides of
+// this with an Arc<TorClientInner>?
 #[derive(Clone)]
 pub struct TorClient<R: Runtime> {
     /// Asynchronous runtime object.
@@ -43,8 +47,16 @@ pub struct TorClient<R: Runtime> {
     circmgr: Arc<tor_circmgr::CircMgr<R>>,
     /// Directory manager for keeping our directory material up to date.
     dirmgr: Arc<tor_dirmgr::DirMgr<R>>,
+    /// Location on disk where we store persistent data.
+    statemgr: FsStateMgr,
     /// Client address configuration
-    addrcfg: ClientAddrConfig,
+    addrcfg: Arc<MutCfg<ClientAddrConfig>>,
+    /// Client DNS configuration
+    timeoutcfg: Arc<MutCfg<StreamTimeoutConfig>>,
+    /// Mutex used to serialize concurrent attempts to reconfigure a TorClient.
+    ///
+    /// See [`TorClient::reconfigure`] for more information on its use.
+    reconfigure_lock: Arc<Mutex<()>>,
 }
 
 /// Preferences for how to route a stream over the Tor network.
@@ -171,6 +183,7 @@ impl<R: Runtime> TorClient<R> {
             );
         }
         let addr_cfg = config.address_filter.clone();
+        let timeout_cfg = config.stream_timeouts.clone();
         let chanmgr = Arc::new(tor_chanmgr::ChanMgr::new(runtime.clone()));
         let circmgr =
             tor_circmgr::CircMgr::new(circ_cfg, statemgr.clone(), &runtime, Arc::clone(&chanmgr))?;
@@ -194,7 +207,7 @@ impl<R: Runtime> TorClient<R> {
         runtime.spawn(update_persistent_state(
             runtime.clone(),
             Arc::downgrade(&circmgr),
-            statemgr,
+            statemgr.clone(),
         ))?;
 
         runtime.spawn(continually_launch_timeout_testing_circuits(
@@ -203,6 +216,12 @@ impl<R: Runtime> TorClient<R> {
             Arc::downgrade(&dirmgr),
         ))?;
 
+        runtime.spawn(continually_preemptively_build_circuits(
+            runtime.clone(),
+            Arc::downgrade(&circmgr),
+            Arc::downgrade(&dirmgr),
+        ))?;
+
         let client_isolation = IsolationToken::new();
 
         Ok(TorClient {
@@ -210,14 +229,82 @@ impl<R: Runtime> TorClient<R> {
             client_isolation,
             circmgr,
             dirmgr,
-            addrcfg: addr_cfg,
+            statemgr,
+            addrcfg: Arc::new(addr_cfg.into()),
+            timeoutcfg: Arc::new(timeout_cfg.into()),
+            reconfigure_lock: Arc::new(Mutex::new(())),
         })
     }
 
-    /// Return a new isolated `TorClient` instance.
+    /// Change the configuration of this TorClient to `new_config`.
     ///
-    /// The two `TorClient`s will share some internal state, but their
-    /// streams will never share circuits with one another.
+    /// The `how` describes whether to perform an all-or-nothing
+    /// reconfiguration: either all of the configuration changes will be
+    /// applied, or none will. If you have disabled all-or-nothing changes, then
+    /// only fatal errors will be reported in this function's return value.
+    ///
+    /// This function applies its changes to **all** TorClient instances derived
+    /// from the same call to [`TorClient::bootstrap`]: even ones whose circuits
+    /// are isolated from this handle.
+    ///
+    /// # Limitations
+    ///
+    /// Although most options are reconfigurable, there are some whose values
+    /// can't be changed on an a running TorClient.  Those options (or their
+    /// sections) are explicitly documented not to be changeable.
+    ///
+    /// Changing some options do not take effect immediately on all open streams
+    /// and circuits, but rather affect only future streams and circuits.  Those
+    /// are also explicitly documented.
+    pub fn reconfigure(
+        &self,
+        new_config: &TorClientConfig,
+        how: tor_config::Reconfigure,
+    ) -> Result<()> {
+        // We need to hold this lock while we're reconfiguring the client: even
+        // though the individual fields have their own synchronization, we can't
+        // safely let two threads change them at once.  If we did, then we'd
+        // introduce time-of-check/time-of-use bugs in checking our configuration,
+        // deciding how to change it, then applying the changes.
+        let _guard = self.reconfigure_lock.lock().expect("Poisoned lock");
+
+        match how {
+            tor_config::Reconfigure::AllOrNothing => {
+                // We have to check before we make any changes.
+                self.reconfigure(new_config, tor_config::Reconfigure::CheckAllOrNothing)?;
+            }
+            tor_config::Reconfigure::CheckAllOrNothing => {}
+            tor_config::Reconfigure::WarnOnFailures => {}
+            _ => {}
+        }
+
+        let circ_cfg = new_config.get_circmgr_config()?;
+        let dir_cfg = new_config.get_dirmgr_config()?;
+        let state_cfg = new_config.storage.expand_state_dir()?;
+        let addr_cfg = &new_config.address_filter;
+        let timeout_cfg = &new_config.stream_timeouts;
+
+        if state_cfg != self.statemgr.path() {
+            how.cannot_change("storage.state_dir")?;
+        }
+
+        self.circmgr.reconfigure(&circ_cfg, how)?;
+        self.dirmgr.reconfigure(&dir_cfg, how)?;
+
+        if how == tor_config::Reconfigure::CheckAllOrNothing {
+            return Ok(());
+        }
+
+        self.addrcfg.replace(addr_cfg.clone());
+        self.timeoutcfg.replace(timeout_cfg.clone());
+
+        Ok(())
+    }
+
+    /// Return a new isolated `TorClient` handle.
+    ///
+    /// The two `TorClient`s will share internal state and configuration, but
+    /// their streams will never share circuits with one another.
     ///
     /// Use this function when you want separate parts of your program to
     /// each have a TorClient handle, but where you don't want their
@@ -226,6 +313,7 @@ impl<R: Runtime> TorClient<R> {
     /// Calling this function is usually preferable to creating a
     /// completely separate TorClient instance, since it can share its
     /// internals with the existing `TorClient`.
+    #[must_use]
     pub fn isolated_client(&self) -> TorClient<R> {
         let mut result = self.clone();
         result.client_isolation = IsolationToken::new();
@@ -237,52 +325,60 @@ impl<R: Runtime> TorClient<R> {
     ///
     /// Note that because Tor prefers to do DNS resolution on the remote
     /// side of the network, this function takes its address as a string.
-    pub async fn connect<A: IntoTorAddr>(
+    pub async fn connect<A: IntoTorAddr>(&self, target: A) -> Result<DataStream> {
+        self.connect_with_prefs(target, ConnectPrefs::default())
+            .await
+    }
+
+    /// Launch an anonymized connection to the provided address and
+    /// port over the Tor network with connection preference flags.
+    ///
+    /// Note that because Tor prefers to do DNS resolution on the remote
+    /// side of the network, this function takes its address as a string.
+    pub async fn connect_with_prefs<A: IntoTorAddr>(
         &self,
         target: A,
-        flags: Option<ConnectPrefs>,
+        flags: ConnectPrefs,
     ) -> Result<DataStream> {
         let addr = target.into_tor_addr()?;
-        addr.enforce_config(&self.addrcfg)?;
+        addr.enforce_config(&self.addrcfg.get())?;
         let (addr, port) = addr.into_string_and_port();
 
-        let flags = flags.unwrap_or_default();
         let exit_ports = [flags.wrap_target_port(port)];
         let circ = self.get_or_launch_exit_circ(&exit_ports, &flags).await?;
         info!("Got a circuit for {}:{}", addr, port);
 
-        // TODO: make this configurable.
-        let stream_timeout = Duration::new(10, 0);
-
         let stream_future = circ.begin_stream(&addr, port, Some(flags.stream_parameters()));
         // This timeout is needless but harmless for optimistic streams.
         let stream = self
             .runtime
-            .timeout(stream_timeout, stream_future)
+            .timeout(self.timeoutcfg.get().connect_timeout, stream_future)
             .await??;
 
         Ok(stream)
     }
 
     /// On success, return a list of IP addresses.
-    pub async fn resolve(
+    pub async fn resolve(&self, hostname: &str) -> Result<Vec<IpAddr>> {
+        self.resolve_with_prefs(hostname, ConnectPrefs::default())
+            .await
+    }
+
+    /// On success, return a list of IP addresses, but use flags.
+    pub async fn resolve_with_prefs(
         &self,
         hostname: &str,
-        flags: Option<ConnectPrefs>,
+        flags: ConnectPrefs,
     ) -> Result<Vec<IpAddr>> {
         let addr = (hostname, 0).into_tor_addr()?;
-        addr.enforce_config(&self.addrcfg)?;
+        addr.enforce_config(&self.addrcfg.get())?;
 
-        let flags = flags.unwrap_or_default();
         let circ = self.get_or_launch_exit_circ(&[], &flags).await?;
 
-        // TODO: make this configurable.
-        let resolve_timeout = Duration::new(10, 0);
-
         let resolve_future = circ.resolve(hostname);
         let addrs = self
             .runtime
-            .timeout(resolve_timeout, resolve_future)
+            .timeout(self.timeoutcfg.get().resolve_timeout, resolve_future)
             .await??;
 
         Ok(addrs)
@@ -291,21 +387,28 @@ impl<R: Runtime> TorClient<R> {
     /// Perform a remote DNS reverse lookup with the provided IP address.
     ///
     /// On success, return a list of hostnames.
-    pub async fn resolve_ptr(
+    pub async fn resolve_ptr(&self, addr: IpAddr) -> Result<Vec<String>> {
+        self.resolve_ptr_with_prefs(addr, ConnectPrefs::default())
+            .await
+    }
+
+    /// Perform a remote DNS reverse lookup with the provided IP address.
+    ///
+    /// On success, return a list of hostnames.
+    pub async fn resolve_ptr_with_prefs(
         &self,
         addr: IpAddr,
-        flags: Option<ConnectPrefs>,
+        flags: ConnectPrefs,
     ) -> Result<Vec<String>> {
-        let flags = flags.unwrap_or_default();
         let circ = self.get_or_launch_exit_circ(&[], &flags).await?;
 
-        // TODO: make this configurable.
-        let resolve_ptr_timeout = Duration::new(10, 0);
-
         let resolve_ptr_future = circ.resolve_ptr(addr);
         let hostnames = self
             .runtime
-            .timeout(resolve_ptr_timeout, resolve_ptr_future)
+            .timeout(
+                self.timeoutcfg.get().resolve_ptr_timeout,
+                resolve_ptr_future,
+            )
             .await??;
 
         Ok(hostnames)
@@ -335,7 +438,7 @@ impl<R: Runtime> TorClient<R> {
         &self,
         exit_ports: &[TargetPort],
         flags: &ConnectPrefs,
-    ) -> Result<Arc<ClientCirc>> {
+    ) -> Result<ClientCirc> {
         let dir = self.dirmgr.netdir();
 
         let isolation = {
@@ -447,7 +550,7 @@ async fn update_persistent_state<R: Runtime>(
             debug!("Circmgr has disappeared; task exiting.");
             return;
         }
-        // XXXX This delay is probably too small.
+        // TODO(nickm): This delay is probably too small.
         //
         // Also, we probably don't even want a fixed delay here.  Instead,
         // we should be updating more frequently when the data is volatile
@@ -476,26 +579,46 @@ async fn continually_launch_timeout_testing_circuits<R: Runtime>(
     circmgr: Weak<tor_circmgr::CircMgr<R>>,
     dirmgr: Weak<tor_dirmgr::DirMgr<R>>,
 ) {
-    loop {
-        let delay;
-        if let (Some(cm), Some(dm)) = (Weak::upgrade(&circmgr), Weak::upgrade(&dirmgr)) {
-            let netdir = dm.netdir();
-            if let Err(e) = cm.launch_timeout_testing_circuit_if_appropriate(&netdir) {
-                warn!("Problem launching a timeout testing circuit: {}", e);
-            }
-            delay = netdir
-                .params()
-                .cbt_testing_delay
-                .try_into()
-                .expect("Out-of-bounds value from BoundedInt32");
-        } else {
-            break;
-        };
+    while let (Some(cm), Some(dm)) = (Weak::upgrade(&circmgr), Weak::upgrade(&dirmgr)) {
+        let netdir = dm.netdir();
+        if let Err(e) = cm.launch_timeout_testing_circuit_if_appropriate(&netdir) {
+            warn!("Problem launching a timeout testing circuit: {}", e);
+        }
+        let delay = netdir
+            .params()
+            .cbt_testing_delay
+            .try_into()
+            .expect("Out-of-bounds value from BoundedInt32");
 
+        drop((cm, dm));
         rt.sleep(delay).await;
     }
 }
 
+/// Run indefinitely, launching circuits where the preemptive circuit
+/// predictor thinks it'd be a good idea to have them.
+///
+/// Exit when we notice that `circmgr` or `dirmgr` has been dropped.
+///
+/// This is a daemon task: it runs indefinitely in the background.
+///
+/// # Note
+///
+/// This would be better handled entirely within `tor-circmgr`, like
+/// other daemon tasks.
+async fn continually_preemptively_build_circuits<R: Runtime>(
+    rt: R,
+    circmgr: Weak<tor_circmgr::CircMgr<R>>,
+    dirmgr: Weak<tor_dirmgr::DirMgr<R>>,
+) {
+    while let (Some(cm), Some(dm)) = (Weak::upgrade(&circmgr), Weak::upgrade(&dirmgr)) {
+        let netdir = dm.netdir();
+        cm.launch_circuits_preemptively(DirInfo::Directory(&netdir))
+            .await;
+        rt.sleep(Duration::from_secs(10)).await;
+    }
+}
+
 impl<R: Runtime> Drop for TorClient<R> {
     // TODO: Consider moving this into tor-circmgr after we have more
     // experience with the state system.
