@@ -7,6 +7,7 @@ use serde::Deserialize;
 use std::collections::HashMap;
 use std::path::Path;
 use std::path::PathBuf;
+use std::time::Duration;
 use tor_config::CfgPath;
 
 pub use tor_config::ConfigBuildError;
@@ -15,7 +16,7 @@ pub use tor_config::ConfigBuildError;
 pub mod circ {
     pub use tor_circmgr::{
         CircMgrConfig, CircMgrConfigBuilder, CircuitTiming, CircuitTimingBuilder, PathConfig,
-        PathConfigBuilder,
+        PathConfigBuilder, PreemptiveCircuitConfig, PreemptiveCircuitConfigBuilder,
     };
 }
 
@@ -32,6 +33,10 @@ pub mod dir {
 ///
 /// This type is immutable once constructed. To create an object of this type,
 /// use [`ClientAddrConfigBuilder`].
+///
+/// You can replace this configuration on a running Arti client.  Doing so will
+/// affect new streams and requests, but will have no effect on existing streams
+/// and requests.
 #[derive(Debug, Clone, Builder, Deserialize, Eq, PartialEq)]
 #[builder(build_fn(error = "ConfigBuildError"))]
 #[serde(deny_unknown_fields)]
@@ -45,6 +50,37 @@ pub struct ClientAddrConfig {
     pub(crate) allow_local_addrs: bool,
 }
 
+/// Configuration for client behavior relating to stream connection timeouts
+///
+/// This type is immutable once constructed. To create an object of this type,
+/// use [`StreamTimeoutConfigBuilder`].
+///
+/// You can replace this configuration on a running Arti client.  Doing so will
+/// affect new streams and requests, but will have no effect on existing streams
+/// and requests—even those that are currently waiting.
+#[derive(Debug, Clone, Builder, Deserialize, Eq, PartialEq)]
+#[builder(build_fn(error = "ConfigBuildError"))]
+#[serde(deny_unknown_fields)]
+#[non_exhaustive]
+pub struct StreamTimeoutConfig {
+    /// How long should we wait before timing out a stream when connecting
+    /// to a host?
+    #[builder(default = "default_connect_timeout()")]
+    #[serde(with = "humantime_serde", default = "default_connect_timeout")]
+    pub(crate) connect_timeout: Duration,
+
+    /// How long should we wait before timing out when resolving a DNS record?
+    #[builder(default = "default_dns_resolve_timeout()")]
+    #[serde(with = "humantime_serde", default = "default_dns_resolve_timeout")]
+    pub(crate) resolve_timeout: Duration,
+
+    /// How long should we wait before timing out when resolving a DNS
+    /// PTR record?
+    #[builder(default = "default_dns_resolve_ptr_timeout()")]
+    #[serde(with = "humantime_serde", default = "default_dns_resolve_ptr_timeout")]
+    pub(crate) resolve_ptr_timeout: Duration,
+}
+
 // NOTE: it seems that `unwrap` may be safe because of builder defaults
 // check `derive_builder` documentation for details
 // https://docs.rs/derive_builder/0.10.2/derive_builder/#default-values
@@ -70,6 +106,47 @@ impl ClientAddrConfig {
     }
 }
 
+#[allow(clippy::unwrap_used)]
+impl Default for StreamTimeoutConfig {
+    fn default() -> Self {
+        StreamTimeoutConfigBuilder::default().build().unwrap()
+    }
+}
+
+impl From<StreamTimeoutConfig> for StreamTimeoutConfigBuilder {
+    fn from(cfg: StreamTimeoutConfig) -> StreamTimeoutConfigBuilder {
+        let mut builder = StreamTimeoutConfigBuilder::default();
+        builder
+            .connect_timeout(cfg.connect_timeout)
+            .resolve_timeout(cfg.resolve_timeout)
+            .resolve_ptr_timeout(cfg.resolve_ptr_timeout);
+
+        builder
+    }
+}
+
+impl StreamTimeoutConfig {
+    /// Return a new [`StreamTimeoutConfigBuilder`].
+    pub fn builder() -> StreamTimeoutConfigBuilder {
+        StreamTimeoutConfigBuilder::default()
+    }
+}
+
+/// Return the default stream timeout
+fn default_connect_timeout() -> Duration {
+    Duration::new(10, 0)
+}
+
+/// Return the default resolve timeout
+fn default_dns_resolve_timeout() -> Duration {
+    Duration::new(10, 0)
+}
+
+/// Return the default PTR resolve timeout
+fn default_dns_resolve_ptr_timeout() -> Duration {
+    Duration::new(10, 0)
+}
+
 /// Configuration for where information should be stored on disk.
 ///
 /// By default, cache information will be stored in `${ARTI_CACHE}`, and
@@ -83,6 +160,8 @@ impl ClientAddrConfig {
 /// information, see the documentation for [`CfgPath`].
 ///
 /// This section is for read/write storage.
+///
+/// You cannot change this section on a running Arti client.
 #[derive(Deserialize, Debug, Clone, Builder, Eq, PartialEq)]
 #[serde(deny_unknown_fields)]
 #[builder(build_fn(error = "ConfigBuildError"))]
@@ -185,11 +264,17 @@ pub struct TorClientConfig {
     /// Information about how to build paths through the network.
     path_rules: circ::PathConfig,
 
+    /// Information about preemptive circuits.
+    preemptive_circuits: circ::PreemptiveCircuitConfig,
+
     /// Information about how to retry and expire circuits and request for circuits.
     circuit_timing: circ::CircuitTiming,
 
     /// Rules about which addresses the client is willing to connect to.
     pub(crate) address_filter: ClientAddrConfig,
+
+    /// Information about timing out client requests.
+    pub(crate) stream_timeouts: StreamTimeoutConfig,
 }
 
 impl Default for TorClientConfig {
@@ -246,9 +331,13 @@ pub struct TorClientConfigBuilder {
     /// Inner builder for the `path_rules` section.
     path_rules: circ::PathConfigBuilder,
     /// Inner builder for the `circuit_timing` section.
+    preemptive_circuits: circ::PreemptiveCircuitConfigBuilder,
+    /// Inner builder for the `circuit_timing` section.
     circuit_timing: circ::CircuitTimingBuilder,
     /// Inner builder for the `address_filter` section.
     address_filter: ClientAddrConfigBuilder,
+    /// Inner builder for the `stream_timeouts` section.
+    stream_timeouts: StreamTimeoutConfigBuilder,
 }
 
 impl TorClientConfigBuilder {
@@ -268,6 +357,10 @@ impl TorClientConfigBuilder {
             .path_rules
             .build()
             .map_err(|e| e.within("path_rules"))?;
+        let preemptive_circuits = self
+            .preemptive_circuits
+            .build()
+            .map_err(|e| e.within("preemptive_circuits"))?;
         let circuit_timing = self
             .circuit_timing
             .build()
@@ -276,6 +369,10 @@ impl TorClientConfigBuilder {
             .address_filter
             .build()
             .map_err(|e| e.within("address_filter"))?;
+        let stream_timeouts = self
+            .stream_timeouts
+            .build()
+            .map_err(|e| e.within("stream_timeouts"))?;
 
         Ok(TorClientConfig {
             tor_network,
@@ -283,8 +380,10 @@ impl TorClientConfigBuilder {
             download_schedule,
             override_net_params,
             path_rules,
+            preemptive_circuits,
             circuit_timing,
             address_filter,
+            stream_timeouts,
         })
     }
 
@@ -356,6 +455,13 @@ impl TorClientConfigBuilder {
         &mut self.path_rules
     }
 
+    /// Return a mutable reference to a [`PreemptiveCircuitConfigBuilder`](circ::PreemptiveCircuitConfigBuilder).
+    ///
+    /// This section overrides Arti's rules for preemptive circuits.
+    pub fn preemptive_circuits(&mut self) -> &mut circ::PreemptiveCircuitConfigBuilder {
+        &mut self.preemptive_circuits
+    }
+
     /// Return a mutable reference to a [`CircuitTimingBuilder`](circ::CircuitTimingBuilder).
     ///
     /// This section overrides Arti's rules for deciding how long to use
@@ -364,6 +470,15 @@ impl TorClientConfigBuilder {
         &mut self.circuit_timing
     }
 
+    /// Return a mutable reference to a [`StreamTimeoutConfigBuilder`].
+    ///
+    /// This section overrides Arti's rules for deciding how long a stream
+    /// request (that is, an attempt to connect or resolve) should wait
+    /// for a response before deciding that the stream has timed out.
+    pub fn stream_timeouts(&mut self) -> &mut StreamTimeoutConfigBuilder {
+        &mut self.stream_timeouts
+    }
+
     /// Return a mutable reference to a [`ClientAddrConfigBuilder`].
     ///
     /// This section controls which addresses Arti is willing to launch connections
@@ -382,8 +497,10 @@ impl From<TorClientConfig> for TorClientConfigBuilder {
             download_schedule,
             override_net_params,
             path_rules,
+            preemptive_circuits,
             circuit_timing,
             address_filter,
+            stream_timeouts,
         } = cfg;
 
         TorClientConfigBuilder {
@@ -392,8 +509,10 @@ impl From<TorClientConfig> for TorClientConfigBuilder {
             download_schedule: download_schedule.into(),
             override_net_params,
             path_rules: path_rules.into(),
+            preemptive_circuits: preemptive_circuits.into(),
             circuit_timing: circuit_timing.into(),
             address_filter: address_filter.into(),
+            stream_timeouts: stream_timeouts.into(),
         }
     }
 }
