@@ -68,7 +68,7 @@
 //! let tor_client = TorClient::bootstrap(rt, config).await?;
 //!
 //! // Initiate a connection over Tor to example.com, port 80.
-//! let mut stream = tor_client.connect(("example.com", 80), None).await?;
+//! let mut stream = tor_client.connect(("example.com", 80)).await?;
 //!
 //! use futures::io::{AsyncReadExt, AsyncWriteExt};
 //!
@@ -144,6 +144,7 @@
 #![deny(missing_docs)]
 #![warn(noop_method_call)]
 #![deny(unreachable_pub)]
+#![deny(clippy::all)]
 #![deny(clippy::await_holding_lock)]
 #![deny(clippy::cargo_common_metadata)]
 #![deny(clippy::cast_lossless)]
@@ -186,5 +187,5 @@ pub use tor_proto::stream::DataStream;
 mod err;
 pub use err::Error;
 
-/// Result type used by this crate
-type Result<T> = std::result::Result<T, Error>;
+/// Alias for the [`Result`] type used by the `arti_client` crate.
+pub type Result<T> = std::result::Result<T, Error>;
